/* bumpref.c
**
**    bummp-reflection-mapping using
**    NVIDIA vertex-shaders register-combiners and texture-programs
**    Copyright (C) 2002  Florian Berger
**    Email: harpin_floh@yahoo.de, florian.berger@jk.uni-linz.ac.at
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License Version 2 as
**    published by the Free Software Foundation;
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
*/

#include <GL/gl.h>
#include <GL/glu.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "bumpref.h"
#include "png_loader.h"
#include "options.h"
#include "vmath.h"


#ifndef NO_NV_BUMPREF  /* force not to use bumpref */

#ifdef GL_NV_register_combiners
#ifdef GL_NV_texture_shader
#ifdef GL_NV_vertex_program

#define USE_NV_BUMPREF

#endif /*GL_NV_vertex_program*/
#endif /*GL_NV_texture_shader*/
#endif /*GL_NV_register_combiners*/

#endif /* NO_NV_BUMPREF */

int bumpref_create_cubemap( char * posx_name,
                           char * posy_name,
                           char * posz_name,
                           char * negx_name,
                           char * negy_name,
                           char * negz_name )
{
#ifdef USE_NV_BUMPREF
    int i, target, cube_tex_bind;
    unsigned char * data;

            DPRINTF("setting environment cube map\n");

            glActiveTextureARB(GL_TEXTURE2_ARB);
            glEnable(GL_TEXTURE_CUBE_MAP_ARB);
//            glEnable(GL_TEXTURE_2D);
            glGenTextures( 1 , &cube_tex_bind );
            glBindTexture(GL_TEXTURE_CUBE_MAP_ARB, cube_tex_bind);
//            glTexParameterf(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
//            glTexParameterf(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
            glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MIN_FILTER, options_tex_min_filter);
            glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MAG_FILTER, options_tex_mag_filter);
            glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
            glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
            glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
//            glBindTexture(GL_TEXTURE_2D, cube_tex_bind);

            for(i=0;i<6;i++){
                int w,h,depth;
/*                w=64; h=64;
                data=malloc(w*h*3);
                for(j=0;j<w*h;j++){
                    if( ((j%w)%9)<1 || ((j/w)%9)<1 ){
                        data[j*3+0]=255*(j%w)/w;
                        data[j*3+1]=255*(j/w)/w;
                        data[j*3+2]=0;
                    }else{
                        data[j*3+0]=((i%3)==0 || i==5)?0xFF:0;
                        data[j*3+1]=((i%3)==1 || i==3)?0xFF:0;
                        data[j*3+2]=((i%3)==2 || i==4)?0xFF:0;
                    }
                }*/
//                                    load_png("chic2.png", &w, &h, &depth, &data);
                DPRINTF("setting cube map #%d\n",i);
                switch(i){
                case 0: load_png(posx_name, &w, &h, &depth, (char **)&data); break;
                case 1: load_png(posy_name, &w, &h, &depth, (char **)&data); break;
                case 2: load_png(posz_name, &w, &h, &depth, (char **)&data); break;
                case 3: load_png(negx_name, &w, &h, &depth, (char **)&data); break;
                case 4: load_png(negy_name, &w, &h, &depth, (char **)&data); break;
                case 5: load_png(negz_name, &w, &h, &depth, (char **)&data); break;
                }
                switch(i){
                case 0: target=GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB; break;
                case 1: target=GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB; break;
                case 2: target=GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB; break;
                case 3: target=GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB; break;
                case 4: target=GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB; break;
                case 5: target=GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB; break;
                }
                DPRINTF("rendering mipmaps for #%d\n",i);
                gluBuild2DMipmaps(target, 3, w, h, GL_RGB, GL_UNSIGNED_BYTE, data);
/*                glTexImage2D(target,
                             0,
                             3,
                             w,
                             h,
                             0,
                             GL_RGB,
                             GL_UNSIGNED_BYTE,
                             data);*/
                DPRINTF("freeing data #%d\n",i);
                free(data);
            }

//            g_cubemap_bind = cube_tex_bind;

            return(cube_tex_bind);
#else
            return(0);
#endif
}



void gaussian_blur( int w, int h, double * ddata, int d )
{
    int xi, yi, xi2, yi2, dx, dy, i;
    double * wdata;
    double gauss_sum, gauss;

    wdata = malloc(sizeof(double)*w*h);

    for(i=0;i<w*h;i++){
        wdata[i]=ddata[i];
    }

    for(yi=0;yi<h;yi++){
        for(xi=0;xi<w;xi++){
            ddata[yi*w+xi]=0.0;
            gauss_sum=0.0;
            for(dy=-d;dy<=d;dy++){
                for(dx=-d;dx<=d;dx++){
                    gauss=exp(-2.0*(double)(dx*dx+dy*dy)/(double)(d*d));
                    xi2=xi+dx; if(xi2<0)xi2+=w; if(xi2>=w)xi2-=w;
                    yi2=yi+dy; if(yi2<0)yi2+=h; if(yi2>=h)yi2-=h;
                    ddata[yi*w+xi]+=gauss*wdata[yi2*w+xi2];
                    gauss_sum+=gauss;
                }
            }
            ddata[yi*w+xi]/=gauss_sum;
        }
    }

    free(wdata);
}

void bump2normal(int w, int h, char * data, double strength)
/* data has to be rgb */
{
    int xi, yi, i, xi_n, yi_n;
    double sd, sd_xn, sd_yn, sd_xyn, dzx, dzy, dx, dy;
    unsigned char * udata;
    double * ddata;
    VMvect n;

    udata = (unsigned char *)data;
    ddata = malloc(sizeof(double)*w*h);

    for(i=0;i<w*h;i++){
        ddata[i]=udata[i*3];
    }
    gaussian_blur( w,h, ddata, 3 );

    for(yi=0;yi<h;yi++){
        for(xi=0;xi<w;xi++){

            i=(yi*w+xi)*3;

            xi_n = (xi+1<w)?xi+1:0;
            yi_n = (yi+1<h)?yi+1:0;

/*            sd     = udata[(yi*w+xi)*3];
            sd_xn  = udata[(yi*w+xi_n)*3];
            sd_yn  = udata[(yi_n*w+xi)*3];
            sd_xyn = udata[(yi_n*w+xi_n)*3];*/

            sd     = ddata[yi*w+xi];
            sd_xn  = ddata[yi*w+xi_n];
            sd_yn  = ddata[yi_n*w+xi];
            sd_xyn = ddata[yi_n*w+xi_n];

            dzx = (sd_xn-sd)*strength;
            dzy = (sd_yn-sd)*strength;

            dx=1.0;
            dy=1.0;

            n=vec_scale(vec_unit(vec_cross(vec_xyz(dx,0,dzx),vec_xyz(0,-dy,dzy))),-1.0);

            udata[i+2]=0x80+127.4*1.0*n.z;
            udata[i+1]=0x80+127.4*1.0*n.y;
            udata[i+0]=0x80+127.4*1.0*n.x;
        }
    }

    free(ddata);
}



signed short * bump2normal_HILO(int w, int h, char * data, double strength)
/* data has to be rgb */
{
    int xi, yi, i, xi_n, yi_n;
    double sd, sd_xn, sd_yn, sd_xyn, dzx, dzy, dx, dy;
    unsigned char * udata;
    signed short * sdata;
    double * ddata;
    VMvect n;

    udata = (unsigned char *)data;
    ddata = malloc(sizeof(double)*w*h);
    sdata = malloc(sizeof(signed short)*w*h*2);

    for(i=0;i<w*h;i++){
        ddata[i]=udata[i*3];
    }
    gaussian_blur( w,h, ddata, 3 );

    for(yi=0;yi<h;yi++){
        for(xi=0;xi<w;xi++){

            i=(yi*w+xi)*3;

            xi_n = (xi+1<w)?xi+1:0;
            yi_n = (yi+1<h)?yi+1:0;

/*            sd     = udata[(yi*w+xi)*3];
            sd_xn  = udata[(yi*w+xi_n)*3];
            sd_yn  = udata[(yi_n*w+xi)*3];
            sd_xyn = udata[(yi_n*w+xi_n)*3];*/

            sd     = ddata[yi*w+xi];
            sd_xn  = ddata[yi*w+xi_n];
            sd_yn  = ddata[yi_n*w+xi];
            sd_xyn = ddata[yi_n*w+xi_n];

            dzx = (sd_xn-sd)*strength;
            dzy = (sd_yn-sd)*strength;

            dx=1.0;
            dy=1.0;

            n=vec_scale(vec_unit(vec_cross(vec_xyz(dx,0,dzx),vec_xyz(0,-dy,dzy))),-1.0);

            sdata[(yi*w+xi)*2+1]=32767.4*n.y;
            sdata[(yi*w+xi)*2+0]=32767.4*n.x;
        }
    }

    free(ddata);

    return(sdata);
}



int bumpref_create_bumpmap( char * map_name, double strength, int hilo )
/* returns texture-bind-id */
{
#ifdef USE_NV_BUMPREF
    int texbind, w, h, depth;
    unsigned char * data;
    signed short * sdata;
    load_png(map_name, &w, &h, &depth, (char **)&data);
    if(hilo){
        sdata = bump2normal_HILO(w, h, data, strength);
    } else {
        bump2normal(w, h, data, strength);
    }

    glGenTextures( 1 , &texbind );
    glBindTexture(GL_TEXTURE_2D,texbind);
    if(!hilo){
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, options_tex_min_filter);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, options_tex_mag_filter);
    } else {
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    }
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    if(!hilo){
        gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB8, w, h, GL_RGB,
                          GL_UNSIGNED_BYTE, data);
/*    glTexImage2D(GL_TEXTURE_2D,
                 0,
                 GL_RGB8,
                 w,
                 h,
                 0,
                 GL_RGB,
                 GL_UNSIGNED_BYTE,
                 sdata);*/
    } else {
        glTexImage2D(GL_TEXTURE_2D,
                     0,
                     GL_SIGNED_HILO16_NV,
                     w,
                     h,
                     0,
                     GL_HILO_NV,
                     GL_SHORT,
                     sdata);
    }

//    g_normalmap_bind = texbind;
    if(hilo){
        free(sdata);
    }

    return(texbind);
#else
    return(0);
#endif
}













void bumpref_setup_vp_ts_rc(
                            char * map_name, double strength,
                            char * posx_name,
                            char * posy_name,
                            char * posz_name,
                            char * negx_name,
                            char * negy_name,
                            char * negz_name,
                            float zoffs, int hilo
                           )
/* dont call this 1st time inside a GL display list */
/* doesnt have any effect when called 1st */
{
#ifdef USE_NV_BUMPREF
    static int init=0;
    static int bumpref_list;
    static int g_cubemap_bind=0;
    static int g_normalmap_bind=0;
    static unsigned int    vert_prog_bind;

    static unsigned char * vert_prog_str=
            "!!VP1.0 \n"
//            #
//            # c[0-3]  = modelview projection (composite) matrix
//            #           TrackMatrixNV(GL_VERTEX_PROGRAM_NV, 0, GL_MODELVIEW_PROJECTION_NV, GL_IDENTITY_NV);
//            # c[4-7]  = modelview transpose
//            #           TrackMatrixNV(GL_VERTEX_PROGRAM_NV, 0, GL_MODELVIEW, GL_INVERSE_TRANSPOSE_NV);
//            # c[8].z  = z offset for z-buffer
//            # c[8].x  = 0.0 ...to circumvent "ADD   R0.z,  R0.z, -R0.z"
//            #
//            # outputs homogenous position and surface basis + eye coord as texture coords
//            #
            /* R2 = mormal vector n */
            "MOV   R2, v[NRML];                      \n"
            /* R0 = xy tangent vector n_xy */
            "MOV   R0.x, -R2.y;                      \n"
            "MOV   R0.y,  R2.x;                      \n"
            "MOV   R0.z,  c[8].x;                    \n"
//            "ADD   R0.z,  R0.z, -R0.z;               \n"
//            "ADD   R0.z,  R0.z, -R0.z;  # R0.z=0 uarrghh \n"
               /* normalize */
            "MOV   R1, R0;                           \n"
            "DP3   R0.w, R1, R1;                     \n"
            "RSQ   R0.w, R0.w;                       \n"
            "MUL   R0.xyz, R1, R0.w;                 \n"
            /* R1 = other tangent vector n_t = n x n_xy */
            "MUL   R1, R2.zxyw, R0.yzxw;             \n"
            "MAD   R1, R2.yzxw, R0.zxyw, -R1;        \n"
            /*transpose of (n_xy, n_t, n) to (tex1, tex2, tex3)*/
            "MOV   o[TEX1].x, R0.x;                  \n"
            "MOV   o[TEX1].y, R1.x;                  \n"
            "MOV   o[TEX1].z, R2.x;                  \n"
            "MOV   o[TEX2].x, R0.y;                  \n"
            "MOV   o[TEX2].y, R1.y;                  \n"
            "MOV   o[TEX2].z, R2.y;                  \n"
            "MOV   o[TEX3].x, R0.z;                  \n"
            "MOV   o[TEX3].y, R1.z;                  \n"
            "MOV   o[TEX3].z, R2.z;                  \n"
            /* dont forget tex0 */
            "MOV   o[TEX0], v[TEX0];                 \n"
            /* R0=camdir=dist(cam,pos) */
            "ADD   R0, c[7], -v[OPOS];               \n"
            /* put this to 4th tex coord */
            "MOV   o[TEX1].w, R0.x;                  \n"
            "MOV   o[TEX2].w, R0.y;                  \n"
            "MOV   o[TEX3].w, R0.z;                  \n"
            /* propagate primary color */
            "MOV   o[COL0], v[COL0];                 \n"
            /* apply modelview + projection and add zbuffer-offset for correct multisampling */
            "DP4   o[HPOS].x, c[0], v[OPOS];         \n"
            "DP4   o[HPOS].y, c[1], v[OPOS];         \n"
            "DP4   R1.z,      c[2], v[OPOS];         \n"
            "DP4   o[HPOS].w, c[3], v[OPOS];         \n"
            /* add zbuffer-offset for correct multisampling */
            "ADD   o[HPOS].z, -c[8].z, R1.z;         \n"
            "END                                     \n";
    GLfloat t_gen_params[] = {1.0,0.0,0.0,0.0};
    GLfloat s_gen_params[] = {0.0,1.0,0.0,0.0};

    if(!init){

        glGenProgramsNV(1, &vert_prog_bind);
        glBindProgramNV(GL_VERTEX_PROGRAM_NV, vert_prog_bind);
        glLoadProgramNV(GL_VERTEX_PROGRAM_NV, vert_prog_bind, strlen(vert_prog_str), vert_prog_str);

#if 1
        bumpref_list = glGenLists(1);
        glNewList(bumpref_list, GL_COMPILE);

        /*vertex program*/
        glTrackMatrixNV(GL_VERTEX_PROGRAM_NV, 0 /* c[0..3] */, GL_MODELVIEW_PROJECTION_NV, GL_IDENTITY_NV);
        glTrackMatrixNV(GL_VERTEX_PROGRAM_NV, 4 /* c[4..7] */, GL_MODELVIEW, GL_INVERSE_TRANSPOSE_NV /*GL_IDENTITY_NV*/);
        glProgramParameter4fNV( GL_VERTEX_PROGRAM_NV, 8, /* c[8] */
                               0.0,     /* needed by vertex prog as constant 0.0 */
                               1.0,     /* dummy - not used */
                               zoffs,   /* z-shift for correct zbuffering when multisampling */
                               1.0      /* dummy - not used */
                              );
        glBindProgramNV(GL_VERTEX_PROGRAM_NV, vert_prog_bind);

        /* texture shader */
        glActiveTextureARB(GL_TEXTURE0_ARB);
        glTexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_TEXTURE_2D);
        glActiveTextureARB(GL_TEXTURE1_ARB);
        glTexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_DOT_PRODUCT_NV);
        if(!hilo) glTexEnvi(GL_TEXTURE_SHADER_NV, GL_RGBA_UNSIGNED_DOT_PRODUCT_MAPPING_NV, GL_EXPAND_NORMAL_NV);
        glTexEnvi(GL_TEXTURE_SHADER_NV, GL_PREVIOUS_TEXTURE_INPUT_NV, GL_TEXTURE0_ARB);
        glActiveTextureARB(GL_TEXTURE2_ARB);
        glTexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_DOT_PRODUCT_NV);
        if(!hilo) glTexEnvi(GL_TEXTURE_SHADER_NV, GL_RGBA_UNSIGNED_DOT_PRODUCT_MAPPING_NV, GL_EXPAND_NORMAL_NV);
        glTexEnvi(GL_TEXTURE_SHADER_NV, GL_PREVIOUS_TEXTURE_INPUT_NV, GL_TEXTURE0_ARB);
        glActiveTextureARB(GL_TEXTURE3_ARB);
        glTexEnvi(GL_TEXTURE_SHADER_NV, GL_SHADER_OPERATION_NV, GL_DOT_PRODUCT_REFLECT_CUBE_MAP_NV);
        if(!hilo) glTexEnvi(GL_TEXTURE_SHADER_NV, GL_RGBA_UNSIGNED_DOT_PRODUCT_MAPPING_NV, GL_EXPAND_NORMAL_NV);
        glTexEnvi(GL_TEXTURE_SHADER_NV, GL_PREVIOUS_TEXTURE_INPUT_NV, GL_TEXTURE0_ARB);
        glActiveTextureARB(GL_TEXTURE0_ARB);

        glCombinerParameteriNV(GL_NUM_GENERAL_COMBINERS_NV,0);
        glFinalCombinerInputNV(GL_VARIABLE_A_NV,
//                               GL_ZERO,
                               GL_TEXTURE3_ARB,
//                               GL_PRIMARY_COLOR_NV,
                               GL_UNSIGNED_IDENTITY_NV,
                               GL_RGB);
        glFinalCombinerInputNV(GL_VARIABLE_B_NV,
//                               GL_ZERO,
                               GL_PRIMARY_COLOR_NV,
                               GL_UNSIGNED_IDENTITY_NV,
                               GL_RGB);
        glFinalCombinerInputNV(GL_VARIABLE_C_NV,
                               GL_ZERO,
                               GL_UNSIGNED_IDENTITY_NV,
                               GL_RGB);
        glFinalCombinerInputNV(GL_VARIABLE_D_NV,
                               GL_ZERO,
//                               GL_TEXTURE3_ARB,
                               GL_UNSIGNED_IDENTITY_NV,
                               GL_RGB);
        glFinalCombinerInputNV(GL_VARIABLE_G_NV,
                               GL_ZERO,
                               GL_UNSIGNED_INVERT_NV,
                               GL_ALPHA);

        glEndList();
#endif

        g_normalmap_bind = bumpref_create_bumpmap( map_name, strength, hilo );
        g_cubemap_bind   = bumpref_create_cubemap( posx_name, posy_name, posz_name, negx_name, negy_name, negz_name );

    }

    if(init){
        glCallList(bumpref_list);

        glEnable(GL_REGISTER_COMBINERS_NV);
        glEnable(GL_TEXTURE_SHADER_NV);
        glEnable(GL_VERTEX_PROGRAM_NV);

        /* normalmap to tex0 */
        glActiveTextureARB(GL_TEXTURE0_ARB);
        //    glEnable(GL_TEXTURE_2D);
        glBindTexture(GL_TEXTURE_2D, g_normalmap_bind);
        /*    glEnable(GL_TEXTURE_GEN_S);
         glEnable(GL_TEXTURE_GEN_T);
         glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
         glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
         glTexGenfv(GL_S, GL_OBJECT_PLANE, s_gen_params);
         glTexGenfv(GL_T, GL_OBJECT_PLANE, t_gen_params);*/

        /* cubemap to tex3 */
        glActiveTextureARB(GL_TEXTURE3_ARB);
        //    glEnable(GL_TEXTURE_CUBE_MAP_ARB);
        glBindTexture(GL_TEXTURE_CUBE_MAP_ARB, g_cubemap_bind);
        /*    glEnable(GL_TEXTURE_GEN_S);
         glEnable(GL_TEXTURE_GEN_T);
         glEnable(GL_TEXTURE_GEN_R);
         glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_ARB);
         glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_ARB);
         glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_ARB);*/

        //    glGetFloatv(GL_MODELVIEW_MATRIX, texmat);
        /* maybe set this per ball to camaera direction */
        //    texmat[12]=0.0;
        //    texmat[13]=0.0;
        //    texmat[14]=0.0;
        /* transpose */
        //    dummy=texmat[1]; texmat[1]=texmat[4]; texmat[4]=dummy;
        //    dummy=texmat[2]; texmat[2]=texmat[8]; texmat[8]=dummy;
        //    du^mmy=texmat[6]; texmat[6]=texmat[9]; texmat[9]=dummy;
        //    glMatrixMode(GL_TEXTURE);
        //    glLoadMatrixf(texmat);
        //    glMatrixMode(GL_MODELVIEW);

        glActiveTextureARB(GL_TEXTURE0_ARB);
    }


    if(!init){
        init=1;
    }
#endif /* ifdef USE_NV_BUMPREF */
}



void bumpref_restore()
{
#ifdef USE_NV_BUMPREF
    glDisable(GL_REGISTER_COMBINERS_NV);
    glDisable(GL_TEXTURE_SHADER_NV);
    glDisable(GL_VERTEX_PROGRAM_NV);

    glActiveTextureARB(GL_TEXTURE3_ARB);
    glDisable(GL_TEXTURE_CUBE_MAP_ARB);
    glDisable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE2_ARB);
    glDisable(GL_TEXTURE_CUBE_MAP_ARB);
    glDisable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glDisable(GL_TEXTURE_CUBE_MAP_ARB);
    glDisable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glDisable(GL_TEXTURE_CUBE_MAP_ARB);
    glEnable(GL_TEXTURE_2D);
/*    glDisable(GL_TEXTURE_GEN_S);
    glDisable(GL_TEXTURE_GEN_T);
    glDisable(GL_TEXTURE_GEN_R);*/

    glActiveTextureARB(GL_TEXTURE3_ARB);
    glDisable(GL_TEXTURE_CUBE_MAP_ARB);
    glDisable(GL_TEXTURE_2D);
    glActiveTextureARB(GL_TEXTURE0_ARB);
/*    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, options_tex_min_filter);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, options_tex_mag_filter);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);*/
#endif
}

